/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              boOX 0-047_dukla                              */
/*                                                                            */
/*                      (C) Copyright 2018 Pavel Surynek                      */
/*                http://www.surynek.com | <pavel@surynek.com>                */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
/* cbs.cpp / 0-047_dukla_047                                                  */
/*----------------------------------------------------------------------------*/
//
// Conflict based search implemented in a standard way. A version for MAPF and
// a version for the token swapping have been implemented.
//
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include <map>

#include "config.h"
#include "compile.h"
#include "version.h"
#include "defs.h"
#include "types.h"
#include "result.h"

#include "core/cbs.h"
#include "util/statistics.h"


using namespace std;
using namespace boOX;


/*----------------------------------------------------------------------------*/

namespace boOX
{


/*----------------------------------------------------------------------------*/
// sCBS

    sCBS::sCBS(sInstance *instance)
	: m_Instance(instance)
    {
	// nothing
    }

    
/*----------------------------------------------------------------------------*/

    sInt_32 sCBS::find_ShortestNonconflictingSwapping(sSolution &solution, sInt_32 cost_limit) const
    {
	return find_ShortestNonconflictingSwapping(*m_Instance, solution, cost_limit);
    }


    sInt_32 sCBS::find_ShortestNonconflictingSwapping(const sInstance &instance, sSolution &solution, sInt_32 cost_limit) const
    {
	sInt_32 cost;
	AgentPaths_vector agent_Paths;

	if ((cost = find_ShortestNonconflictingSwapping(agent_Paths, cost_limit)) < 0)
	{
	    return -1;
	}
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();

	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    #ifdef sVERBOSE
	    {
		printf("Agent %d: ", agent_id);
	    }
	    #endif
	    for (sInt_32 i = 1; i < agent_Paths[agent_id].size(); ++i)
	    {
                #ifdef sVERBOSE
		{
		    printf("%d ", agent_Paths[agent_id][i - 1]);
		}
                #endif
		if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i])
		{
		    solution.add_Move(i - 1, sSolution::Move(agent_id, agent_Paths[agent_id][i - 1], agent_Paths[agent_id][i]));
		}
	    }
            #ifdef sVERBOSE
	    {
		printf("%d\n", *agent_Paths[agent_id].rbegin());
	    }
            #endif	    
	}	
	return cost;
    }

    
    sInt_32 sCBS::find_ShortestNonconflictingSwapping(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	return find_ShortestNonconflictingSwapping(*m_Instance, agent_Paths, cost_limit);
    }

    
    sInt_32 sCBS::find_ShortestNonconflictingSwapping(const sInstance &instance, AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	sInt_32 solution_cost;
	sDouble start_time = sStatistics::get_CPU_Seconds();
	
	for (sInt_32 cost = 0; cost <= cost_limit; ++cost)
	{
	    #ifdef sVERBOSE
	    {
		sDouble end_time = sStatistics::get_CPU_Seconds();
		printf("Solving cost %d (elapsed time [seconds]: %.3f)...\n", cost, (end_time - start_time));
	    }
	    #endif
	    if ((solution_cost = find_NonconflictingSwapping(instance, agent_Paths, cost)) >= 0)
	    {
		return solution_cost;
	    }
	}
	return -1;
    }

    
    sInt_32 sCBS::find_ShortestNonconflictingPaths(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	return find_ShortestNonconflictingPaths(*m_Instance, agent_Paths, cost_limit);
    }

    
    sInt_32 sCBS::find_ShortestNonconflictingPaths(const sInstance &instance, AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	sInt_32 solution_cost;
	
	for (sInt_32 cost = 0; cost <= cost_limit; ++cost)
	{
	    #ifdef sVERBOSE
	    {
		printf("Solving cost %d ...\n", cost);
	    }
	    #endif	    
	    if ((solution_cost = find_NonconflictingPaths(instance, agent_Paths, cost)) >= 0)
	    {
		return solution_cost;
	    }
	}
	return -1;
    }


    sInt_32 sCBS::find_NonconflictingSwapping(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	return find_NonconflictingSwapping(*m_Instance, agent_Paths, cost_limit);
    }
    
    
    sInt_32 sCBS::find_NonconflictingSwapping(const sInstance  &instance, AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	AgentConflicts_vector agent_Conflicts;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();

	agent_Conflicts.resize(N_agents + 1);
	return find_NonconflictingSwapping_prio(instance, agent_Conflicts, agent_Paths, cost_limit);
    }

    
    sInt_32 sCBS::find_NonconflictingPaths(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	return find_NonconflictingPaths(*m_Instance, agent_Paths, cost_limit);
    }
    
    
    sInt_32 sCBS::find_NonconflictingPaths(const sInstance  &instance, AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const
    {
	AgentConflicts_vector agent_Conflicts;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();

	agent_Conflicts.resize(N_agents + 1);
	return find_NonconflictingPaths(instance, agent_Conflicts, agent_Paths, cost_limit);
    }


    #define sCBS_ADD_AGENT_CONFLICT(agent_id, step, vertex_id)    \
    {                                                             \
	if (agent_Conflicts[(agent_id)].size() <= (step))         \
	{                                                         \
	    agent_Conflicts[(agent_id)].resize((step) + 1);       \
	}                                                         \
	agent_Conflicts[(agent_id)][(step)].insert((vertex_id));  \
    }

    
    #define sCBS_DEL_AGENT_CONFLICT(agent_id, step, vertex_id)    \
    {                                                             \
	agent_Conflicts[(agent_id)][(step)].erase((vertex_id));   \
    }


    #define sCBS_ADD_NODE_AGENT_CONFLICT(node, agent_id, step, vertex_id)  \
    {                                                                      \
	if ((node).m_agent_Conflicts[(agent_id)].size() <= (step))         \
	{                                                                  \
	    (node).m_agent_Conflicts[(agent_id)].resize((step) + 1);       \
	}                                                                  \
	(node).m_agent_Conflicts[(agent_id)][(step)].insert((vertex_id));  \
    }

    
    #define sCBS_DEL_NODE_AGENT_CONFLICT(node, agent_id, step, vertex_id)  \
    {                                                                      \
	node.m_agent_Conflicts[(agent_id)][(step)].erase((vertex_id));     \
    }    


    #define sCBS_SHOW_CONFLICTS                                                                          \
    {                                                                                                    \
	for (sInt_32 agent_id = 1; agent_id < agent_Conflicts.size(); ++agent_id)                        \
	{                                                                                                \
	    printf("Agent: %d\n", agent_id);                                                             \
	    sInt_32 N_steps = agent_Conflicts[agent_id].size();	                                         \
	    for (sInt_32 step = 0; step < N_steps; ++step)                                               \
	    {                                                                                            \
		printf("  %d: ", step);                                                                  \
		for (VertexIDs_uset::const_iterator conflict = agent_Conflicts[agent_id][step].begin();  \
		     conflict != agent_Conflicts[agent_id][step].end(); ++conflict)                      \
		{                                                                                        \
		    printf("%d ", *conflict);                                                            \
		}                                                                                        \
		printf("\n");                                                                            \
	    }                                                                                            \
	}                                                                                                \
    }


    sInt_32 sCBS::find_NonconflictingSwapping_base(const sInstance       &instance,
						   AgentConflicts_vector &agent_Conflicts,
						   AgentPaths_vector     &agent_Paths,
						   sInt_32                cost_limit) const
    {
	sInt_32 cost, cummulative, agent_path_length;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();
	
	agent_Paths.clear();
	agent_Paths.resize(N_agents + 1);
	
	Occupations_vector occupations;
	cummulative = 0;

	#ifdef sSTATISTICS
	{
	    ++s_GlobalStatistics.get_CurrentPhase().m_search_Steps;
	}
	#endif

	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    if ((agent_path_length = find_NonconflictingSequence(instance.m_environment,
								 instance.m_start_configuration.get_AgentLocation(agent_id),
								 instance.m_goal_configuration.get_AgentLocation(agent_id),
								 agent_Conflicts[agent_id],
								 agent_Paths[agent_id])) < 0)
	    {
		return -1;
	    }	    
	    if (!occupations.empty())
	    {
		while (occupations.size() < agent_path_length)
		{
		    occupations.push_back(occupations.back());
		}
	    }
	    else
	    {
		occupations.resize(agent_path_length);		
	    }
	    occupations[0][agent_Paths[agent_id][0]] = agent_id;
	    cummulative += agent_path_length;
	    
	    if (cummulative > cost_limit)
	    {
		return -1;
	    }
	}

	for (sInt_32 i = 1;; ++i)
	{
	    bool finished = true;

	    for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	    {
		AgentPaths_vector next_agent_Paths;
		agent_path_length = agent_Paths[agent_id].size();

		if (i < agent_path_length)
		{
		    finished = false;
		    Occupation_umap::const_iterator occupation_collision = occupations[i].find(agent_Paths[agent_id][i]);
		    
		    if (occupation_collision != occupations[i].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingSwapping_base(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			
			sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingSwapping_base(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			return -1;
		    }

		    if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i]) // proper swap
		    {
			Occupation_umap::const_iterator swap_expectation_pred = occupations[i - 1].find(agent_Paths[agent_id][i]);		    
			
			if (swap_expectation_pred != occupations[i - 1].end()) // swap with occupied
			{
			    sInt_32 ii = sMIN(agent_Paths[swap_expectation_pred->second].size() - 1, i);
			    
			    if (agent_Paths[swap_expectation_pred->second][ii] != agent_Paths[agent_id][i - 1])
			    {
				sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
				if ((cost = find_NonconflictingSwapping_base(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
				{
				    agent_Paths = next_agent_Paths;
				    return cost;
				}
				sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
				
				sCBS_ADD_AGENT_CONFLICT(swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				if ((cost = find_NonconflictingSwapping_base(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
				{
				    agent_Paths = next_agent_Paths;
				    return cost;
				}
				sCBS_DEL_AGENT_CONFLICT(swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				return -1;
			    }
			}
		    }
		    occupations[i][agent_Paths[agent_id][i]] = agent_id;
		}
	    }
	    if (finished)
	    {
		break;
	    }
	}
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    agent_path_length = agent_Paths[agent_id].size();
	    sASSERT(agent_path_length > 0);

	    for (sInt_32 i = agent_path_length; i < occupations.size(); ++i)
	    {
		AgentPaths_vector next_agent_Paths;
		Occupation_umap::const_iterator occupation_collision = occupations[i].find(agent_Paths[agent_id][agent_path_length - 1]);

		if (occupation_collision != occupations[i].end())
		{
		    sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = find_NonconflictingSwapping_base(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
		    {
			agent_Paths = next_agent_Paths;
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
			
		    sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = find_NonconflictingSwapping_base(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
		    {
			agent_Paths = next_agent_Paths;
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    return -1;
		}
		occupations[i][agent_Paths[agent_id][agent_path_length - 1]] = agent_id;
	    }
	}
	return cummulative;
    }


    sInt_32 sCBS::find_NonconflictingSwapping_upd(const sInstance       &instance,
						  AgentConflicts_vector &agent_Conflicts,
						  AgentPaths_vector     &agent_Paths,
						  sInt_32                cost_limit) const
    {
	sInt_32 cost, cummulative, agent_path_length;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();
	
	agent_Paths.clear();
	agent_Paths.resize(N_agents + 1);
	
	Occupations_vector space_Occupations;
	cummulative = 0;

	#ifdef sSTATISTICS
	{
	    ++s_GlobalStatistics.get_CurrentPhase().m_search_Steps;
	}
	#endif

	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{	
	    if ((agent_path_length = find_NonconflictingSequence(instance.m_environment,
								 instance.m_start_configuration.get_AgentLocation(agent_id),
								 instance.m_goal_configuration.get_AgentLocation(agent_id),
								 agent_Conflicts[agent_id],
								 agent_Paths[agent_id])) < 0)
	    {
		return -1;
	    }	    
	    if (!space_Occupations.empty())
	    {
		while (space_Occupations.size() < agent_path_length)
		{
		    space_Occupations.push_back(space_Occupations.back());
		}
	    }
	    else
	    {
		space_Occupations.resize(agent_path_length);		
	    }
	    space_Occupations[0][agent_Paths[agent_id][0]] = agent_id;
	    cummulative += agent_path_length;
	    
	    if (cummulative > cost_limit)
	    {
		return -1;
	    }			
	}

	for (sInt_32 i = 1;; ++i)
	{
	    bool finished = true;

	    for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	    {
		agent_path_length = agent_Paths[agent_id].size();

		if (i < agent_path_length)
		{
		    finished = false;
		    Occupation_umap::const_iterator occupation_collision = space_Occupations[i].find(agent_Paths[agent_id][i]);
		    
		    if (occupation_collision != space_Occupations[i].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
			{
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
		    
			sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			if ((cost = update_NonconflictingSwapping(occupation_collision->second, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
			{
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			return -1;
		    }
		    if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i]) // proper swap
		    {
			Occupation_umap::const_iterator swap_expectation_pred = space_Occupations[i - 1].find(agent_Paths[agent_id][i]);		    
			
			if (swap_expectation_pred != space_Occupations[i - 1].end()) // swap with occupied
			{
			    sInt_32 ii = sMIN(agent_Paths[swap_expectation_pred->second].size() - 1, i);
			    
			    if (agent_Paths[swap_expectation_pred->second][ii] != agent_Paths[agent_id][i - 1])
			    {
				sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
				if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
				{
				    return cost;
				}
				sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
				
				sCBS_ADD_AGENT_CONFLICT(swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				if ((cost = update_NonconflictingSwapping(swap_expectation_pred->second, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
				{
				    return cost;
				}
				sCBS_DEL_AGENT_CONFLICT(swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				return -1;
			    }
			}
		    }
/*		    
		    if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i]) // proper swap
		    {
			Occupation_umap::const_iterator swap_expectation_pred = space_Occupations[i - 1].find(agent_Paths[agent_id][i]);		    
			Occupation_umap::const_iterator swap_expectation_post = space_Occupations[i].find(agent_Paths[agent_id][i - 1]);
			
			if (swap_expectation_pred != space_Occupations[i - 1].end() && swap_expectation_post != space_Occupations[i].end() && swap_expectation_pred->second != swap_expectation_post->second)
			{
			    sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			    if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
			    {
				return cost;
			    }
			    sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			    
			    sCBS_ADD_AGENT_CONFLICT(swap_expectation_post->second, i, agent_Paths[agent_id][i - 1]);
			    if ((cost = update_NonconflictingSwapping(swap_expectation_post->second, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
			    {
				return cost;
			    }
			    sCBS_DEL_AGENT_CONFLICT(swap_expectation_post->second, i, agent_Paths[agent_id][i - 1]);
			    return -1;
			}
		    }		
*/
		    space_Occupations[i][agent_Paths[agent_id][i]] = agent_id;
		}
	    }
	    if (finished)
	    {
		break;
	    }
	}

	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    agent_path_length = agent_Paths[agent_id].size();
	    sASSERT(agent_path_length > 0);
	    
	    for (sInt_32 i = agent_path_length; i < space_Occupations.size(); ++i)
	    {
		Occupation_umap::const_iterator occupation_collision = space_Occupations[i].find(agent_Paths[agent_id][agent_path_length - 1]);
		if (occupation_collision != space_Occupations[i].end())
		{
		    sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
		    {
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
			
		    sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
		    {
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    return -1;
		}
		space_Occupations[i][agent_Paths[agent_id][agent_path_length - 1]] = agent_id;
	    }
	}
	return cummulative;
    }


    sInt_32 sCBS::find_NonconflictingSwapping_prio(const sInstance       &instance,
						   AgentConflicts_vector &agent_Conflicts,
						   AgentPaths_vector     &agent_Paths,
						   sInt_32                cost_limit) const
    {
	sInt_32 cummulative, agent_path_length;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();	       
	
	Node initial_node(-1);
	sInt_32 initial_cost = 0;

	initial_node.m_agent_Conflicts = agent_Conflicts;
	initial_node.m_agent_Paths.clear();
	initial_node.m_agent_Paths.resize(N_agents + 1);
	
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    if ((agent_path_length = find_NonconflictingSequence(instance.m_environment,
								 instance.m_start_configuration.get_AgentLocation(agent_id),
								 instance.m_goal_configuration.get_AgentLocation(agent_id),
								 initial_node.m_agent_Conflicts[agent_id],
								 initial_node.m_agent_Paths[agent_id])) < 0)
	    {
		return -1;
	    }
	    initial_cost += agent_path_length;
	}

	Nodes_mmap search_Queue;
	search_Queue.insert(Nodes_mmap::value_type(initial_cost, initial_node));

	while (!search_Queue.empty())
	{
	    /*
	    for (Nodes_mmap::const_iterator nn = search_Queue.begin(); nn != search_Queue.end(); ++nn)
	    {
		printf("%d ", nn->first);
	    }
	    printf("\n");
	    */
            #ifdef sSTATISTICS
	    {
		++s_GlobalStatistics.get_CurrentPhase().m_search_Steps;
	    }
            #endif
	    
	    Node &best_node = search_Queue.begin()->second;

	    if (best_node.m_upd_agent_id > 0)
	    {
		if ((agent_path_length = find_NonconflictingSequence(instance.m_environment,
								     instance.m_start_configuration.get_AgentLocation(best_node.m_upd_agent_id),
								     instance.m_goal_configuration.get_AgentLocation(best_node.m_upd_agent_id),
								     best_node.m_agent_Conflicts[best_node.m_upd_agent_id],
								     best_node.m_agent_Paths[best_node.m_upd_agent_id])) < 0)
		{
		    search_Queue.erase(search_Queue.begin());
		    continue;
		}
	    }

	    if ((cummulative = revise_NonconflictingSwapping(instance,
							     best_node.m_agent_Conflicts,
							     best_node.m_agent_Paths,
							     cost_limit,
							     search_Queue)) < 0)
	    {
		search_Queue.erase(search_Queue.begin());
		continue;
	    }
	    agent_Paths = best_node.m_agent_Paths;
	    return cummulative;
	}
	return -1;
    }


    sInt_32 sCBS::revise_NonconflictingSwapping(const sInstance         &instance,
						AgentConflicts_vector   &agent_Conflicts,
						const AgentPaths_vector &agent_Paths,
						sInt_32                  cost_limit,
						Nodes_mmap              &search_Queue) const
    {
	sInt_32 agent_path_length;
	Occupations_vector space_Occupations;
	sInt_32 cummulative = 0;

	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();	
	
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    agent_path_length = agent_Paths[agent_id].size();

	    if (!space_Occupations.empty())
	    {
		while (space_Occupations.size() < agent_path_length)
		{
		    space_Occupations.push_back(space_Occupations.back());
		}
	    }
	    else
	    {
		space_Occupations.resize(agent_path_length);		
	    }
	    space_Occupations[0][agent_Paths[agent_id][0]] = agent_id;
	    cummulative += agent_path_length;
	    
	    if (cummulative > cost_limit)
	    {
		return -1;
	    }
	}		       
	
	for (sInt_32 i = 1;; ++i)
	{
	    bool finished = true;
		
	    for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	    {
		agent_path_length = agent_Paths[agent_id].size();
		
		if (i < agent_path_length)
		{
		    finished = false;
		    Occupation_umap::const_iterator occupation_collision = space_Occupations[i].find(agent_Paths[agent_id][i]);
		    
		    if (occupation_collision != space_Occupations[i].end())
		    {
			Node next_node(agent_id, agent_Conflicts, agent_Paths);
			sCBS_ADD_NODE_AGENT_CONFLICT(next_node, agent_id, i, agent_Paths[agent_id][i]);
			search_Queue.insert(Nodes_mmap::value_type(cummulative, next_node));
			
			next_node = Node(occupation_collision->second, agent_Conflicts, agent_Paths);
			sCBS_ADD_NODE_AGENT_CONFLICT(next_node, occupation_collision->second, i, agent_Paths[agent_id][i]);
			search_Queue.insert(Nodes_mmap::value_type(cummulative, next_node));

			return -1;			
		    }
		    if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i]) // proper swap
		    {
			Occupation_umap::const_iterator swap_expectation_pred = space_Occupations[i - 1].find(agent_Paths[agent_id][i]);		    
			
			if (swap_expectation_pred != space_Occupations[i - 1].end()) // swap with occupied
			{
			    sInt_32 ii = sMIN(agent_Paths[swap_expectation_pred->second].size() - 1, i);
			    
			    if (agent_Paths[swap_expectation_pred->second][ii] != agent_Paths[agent_id][i - 1])
			    {
				Node next_node(agent_id, agent_Conflicts, agent_Paths);
				sCBS_ADD_NODE_AGENT_CONFLICT(next_node, agent_id, i, agent_Paths[agent_id][i]);
				search_Queue.insert(Nodes_mmap::value_type(cummulative, next_node));
				
				next_node = Node(swap_expectation_pred->second, agent_Conflicts, agent_Paths);
				sCBS_ADD_NODE_AGENT_CONFLICT(next_node, swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				search_Queue.insert(Nodes_mmap::value_type(cummulative, next_node));

				return -1;
			    }
			}
		    }
		    space_Occupations[i][agent_Paths[agent_id][i]] = agent_id;
		}
	    }
	    if (finished)
	    {
		break;
	    }
	}
	
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    agent_path_length = agent_Paths[agent_id].size();
	    sASSERT(agent_path_length > 0);
	    
	    for (sInt_32 i = agent_path_length; i < space_Occupations.size(); ++i)
	    {
		Occupation_umap::const_iterator occupation_collision = space_Occupations[i].find(agent_Paths[agent_id][agent_path_length - 1]);
		if (occupation_collision != space_Occupations[i].end())
		{
		    Node next_node(agent_id, agent_Conflicts, agent_Paths);
		    sCBS_ADD_NODE_AGENT_CONFLICT(next_node, agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
		    search_Queue.insert(Nodes_mmap::value_type(cummulative, next_node));
		    
		    next_node = Node(occupation_collision->second, agent_Conflicts, agent_Paths);
		    sCBS_ADD_NODE_AGENT_CONFLICT(next_node, occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    search_Queue.insert(Nodes_mmap::value_type(cummulative, next_node));

		    return -1;
		}
		space_Occupations[i][agent_Paths[agent_id][agent_path_length - 1]] = agent_id;
	    }
	}
	return cummulative;
    }


    sInt_32 sCBS::update_NonconflictingSwapping(sInt_32                upd_agent_id,
						const sInstance       &instance,
						Occupations_vector    &sUNUSED(space_Occupations_),
						AgentConflicts_vector &agent_Conflicts,
						AgentPaths_vector     &agent_Paths,
						sInt_32                cost_limit) const
    {
	sInt_32 cost, cummulative, agent_path_length;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();
	
	VertexIDs_vector stored_path = agent_Paths[upd_agent_id];
	agent_Paths[upd_agent_id].clear();

	Occupations_vector space_Occupations;
	cummulative = 0;

	#ifdef sSTATISTICS
	{
	    ++s_GlobalStatistics.get_CurrentPhase().m_search_Steps;
	}
	#endif

	{	
	    if ((agent_path_length = find_NonconflictingSequence(instance.m_environment,
								 instance.m_start_configuration.get_AgentLocation(upd_agent_id),
								 instance.m_goal_configuration.get_AgentLocation(upd_agent_id),
								 agent_Conflicts[upd_agent_id],
								 agent_Paths[upd_agent_id])) < 0)
	    {
		agent_Paths[upd_agent_id] = stored_path;
		return -1;
	    }
	}
	
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    agent_path_length = agent_Paths[agent_id].size();
	    
	    if (!space_Occupations.empty())
	    {
		while (space_Occupations.size() < agent_path_length)
		{
		    space_Occupations.push_back(space_Occupations.back());
		}
	    }
	    else
	    {
		space_Occupations.resize(agent_path_length);
	    }
	    cummulative += agent_path_length;

	    space_Occupations[0][agent_Paths[agent_id][0]] = agent_id;
	}
	if (cummulative > cost_limit)
	{
	    agent_Paths[upd_agent_id] = stored_path;
	    return -1;
	}
	
	for (sInt_32 i = 1;; ++i)
	{
	    bool finished = true;
	    
	    for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	    {
		agent_path_length = agent_Paths[agent_id].size();
		
		if (i < agent_path_length)
		{
		    finished = false;
		    Occupation_umap::const_iterator occupation_collision = space_Occupations[i].find(agent_Paths[agent_id][i]);
	
		    if (occupation_collision != space_Occupations[i].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
			{
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
		    
			sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			if ((cost = update_NonconflictingSwapping(occupation_collision->second, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
			{
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			agent_Paths[upd_agent_id] = stored_path;
			return -1;
		    }

		    if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i]) // proper swap
		    {
			Occupation_umap::const_iterator swap_expectation_pred = space_Occupations[i - 1].find(agent_Paths[agent_id][i]);		    
			
			if (swap_expectation_pred != space_Occupations[i - 1].end()) // swap with occupied
			{
			    sInt_32 ii = sMIN(agent_Paths[swap_expectation_pred->second].size() - 1, i);
			    
			    if (agent_Paths[swap_expectation_pred->second][ii] != agent_Paths[agent_id][i - 1])
			    {
				sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
				if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
				{
				    return cost;
				}
				sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
				
				sCBS_ADD_AGENT_CONFLICT(swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				if ((cost = update_NonconflictingSwapping(swap_expectation_pred->second, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
				{
				    return cost;
				}
				sCBS_DEL_AGENT_CONFLICT(swap_expectation_pred->second, ii, agent_Paths[swap_expectation_pred->second][ii]);
				agent_Paths[upd_agent_id] = stored_path;
				return -1;
			    }
			}
		    }
		    space_Occupations[i][agent_Paths[agent_id][i]] = agent_id;
		}
	    }
	    if (finished)
	    {
		break;
	    }
	}
	
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    agent_path_length = agent_Paths[agent_id].size();
	    sASSERT(agent_path_length > 0);
		    
	    for (sInt_32 i = agent_path_length; i < space_Occupations.size(); ++i)
	    {
		/*
		if (agent_path_length <= 0)
		{
		    for (sInt_32 ia = 1; ia < N_agents; ++ia)
		    {
			for (sInt_32 ip = 0; ip < agent_Paths[ia].size(); ++ip)
			{
			    printf("%d ", agent_Paths[ia][ip]);
			}
			printf("\n");
		    }
		}
		*/		    
		
		Occupation_umap::const_iterator occupation_collision = space_Occupations[i].find(agent_Paths[agent_id][agent_path_length - 1]);
		
		if (occupation_collision != space_Occupations[i].end())
		{
		    sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = update_NonconflictingSwapping(agent_id, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
		    {
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
		    
		    sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = update_NonconflictingSwapping(occupation_collision->second, instance, space_Occupations, agent_Conflicts, agent_Paths, cost_limit)) >= 0)
		    {
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    agent_Paths[upd_agent_id] = stored_path;
		    return -1;
		}
		space_Occupations[i][agent_Paths[agent_id][agent_path_length - 1]] = agent_id;
	    }	    
	}
	return cummulative;
    }        


    sInt_32 sCBS::find_NonconflictingPaths(const sInstance       &instance,
					   AgentConflicts_vector &agent_Conflicts,
					   AgentPaths_vector     &agent_Paths,
					   sInt_32                cost_limit) const
    {
	sInt_32 cost, cummulative, agent_path_length;
	sInt_32 N_agents = instance.m_start_configuration.get_AgentCount();
	
	agent_Paths.clear();
	agent_Paths.resize(N_agents + 1);
	
	Occupations_vector occupations;
	cummulative = 0;

	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    AgentPaths_vector next_agent_Paths;
		
	    if ((agent_path_length = find_NonconflictingSequence(instance.m_environment,
								 instance.m_start_configuration.get_AgentLocation(agent_id),
								 instance.m_goal_configuration.get_AgentLocation(agent_id),
								 agent_Conflicts[agent_id],
								 agent_Paths[agent_id])) < 0)
	    {
		return -1;
	    }
	    if (!occupations.empty())
	    {
		while (occupations.size() < agent_path_length)
		{
		    occupations.push_back(occupations.back());
		}
	    }
	    else
	    {
		occupations.resize(agent_path_length);		
	    }
	    occupations[0][agent_Paths[agent_id][0]] = agent_id;

	    cummulative += agent_path_length;
	    
	    if (cummulative > cost_limit)
	    {
		return -1;
	    }			
	}
	for (sInt_32 agent_id = 1; agent_id <= N_agents; ++agent_id)
	{
	    AgentPaths_vector next_agent_Paths;
	    agent_path_length = agent_Paths[agent_id].size();
			
	    for (sInt_32 i = 1; i < agent_path_length; ++i)
	    {
		sCBS_SHOW_CONFLICTS;
		
		if (agent_Paths[agent_id][i - 1] != agent_Paths[agent_id][i]) /* non-trivial movement */
		{
		    Occupation_umap::const_iterator occupation_vacant = occupations[i - 1].find(agent_Paths[agent_id][i]);
		    if (occupation_vacant != occupations[i - 1].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			
			sCBS_ADD_AGENT_CONFLICT(occupation_vacant->second, i - 1, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(occupation_vacant->second, i - 1, agent_Paths[agent_id][i]);			
			return -1;
		    }
		    Occupation_umap::const_iterator occupation_left = occupations[i].find(agent_Paths[agent_id][i - 1]);
		    if (occupation_left != occupations[i].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i - 1, agent_Paths[agent_id][i - 1]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i - 1, agent_Paths[agent_id][i - 1]);

			sCBS_ADD_AGENT_CONFLICT(occupation_left->second, i, agent_Paths[agent_id][i - 1]);			
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(occupation_left->second, i, agent_Paths[agent_id][i - 1]);
			return -1;
		    }

		    Occupation_umap::const_iterator occupation_collision = occupations[i].find(agent_Paths[agent_id][i]);
		    if (occupation_collision != occupations[i].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);

			sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			return -1;
		    }
		}
		else /* wait action */
		{
		    Occupation_umap::const_iterator occupation_collision = occupations[i].find(agent_Paths[agent_id][i]);
		    if (occupation_collision != occupations[i].end())
		    {
			sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}
			sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][i]);
			
			sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
			{
			    agent_Paths = next_agent_Paths;
			    return cost;
			}			
			sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][i]);
			return -1;
		    }
		}
		if (++cummulative > cost_limit)
		{
		    return -1;
		}
		occupations[i][agent_Paths[agent_id][i]] = agent_id;
	    }
	    
	    for (sInt_32 i = agent_path_length; i < occupations.size(); ++i)
	    {
		Occupation_umap::const_iterator occupation_collision = occupations[i].find(agent_Paths[agent_id][agent_path_length - 1]);
		if (occupation_collision != occupations[i].end())
		{
		    sCBS_ADD_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);
		    if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
		    {
			agent_Paths = next_agent_Paths;
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(agent_id, i, agent_Paths[agent_id][agent_path_length - 1]);

		    sCBS_ADD_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);			
		    if ((cost = find_NonconflictingPaths(instance, agent_Conflicts, next_agent_Paths, cost_limit)) >= 0)
		    {
			agent_Paths = next_agent_Paths;
			return cost;
		    }
		    sCBS_DEL_AGENT_CONFLICT(occupation_collision->second, i, agent_Paths[agent_id][agent_path_length - 1]);
		    return -1;
		}
		occupations[i][agent_Paths[agent_id][agent_path_length - 1]] = agent_id;
	    }
	    
	}
	return cummulative;
    }

    
/*----------------------------------------------------------------------------*/
    
    sInt_32 sCBS::find_NonconflictingSequence(const sUndirectedGraph &graph,
					      sInt_32                 source_id,
					      sInt_32                 sink_id,
					      const Conflicts_vector &Conflicts,
					      VertexIDs_vector       &Path) const
    {
	Visits_list visit_Queue;
	Visits_vector visited_Vertices;
	
	/*	
	for (sInt_32 level = 0; level < Conflicts.size(); ++level)
	{
	    printf("%d: ", level);
	    for (VertexIDs_uset::const_iterator conf = Conflicts[level].begin(); conf != Conflicts[level].end(); ++conf)
	    {
		printf("%d ", *conf);
	    }
	    printf("\n");
	}	
	*/
	if (source_id == sink_id)
	{
	    bool non_conflicting_sink = true;

	    for (sInt_32 level = 0; level < Conflicts.size(); ++level)
	    {
		if (Conflicts[level].find(sink_id) != Conflicts[level].end())
		{
		    non_conflicting_sink = false;
		    break;
		}
	    }

	    if (non_conflicting_sink)
	    {
		Path.push_back(sink_id);
		return 1;
	    }
	}
  
	if (Conflicts.empty() || Conflicts[0].find(source_id) == Conflicts[0].end())
	{
	    visit_Queue.push_back(Visit(0, source_id, -1));
	    visited_Vertices.push_back(Visits_umap());
	    
	    Visit source_visit(0, source_id, -1);
	    visited_Vertices[0][source_id] = source_visit;
	}
	else
	{
	    return -1;
	}
	
	while (!visit_Queue.empty())
	{
	    const Visit &front_visit = visit_Queue.front();
	    if (visited_Vertices.size() <= front_visit.m_level + 1)
	    {
		visited_Vertices.push_back(Visits_umap());
	    }
	    for (sVertex::Neighbors_list::const_iterator neighbor = graph.m_Vertices[front_visit.m_vertex_id].m_Neighbors.begin(); neighbor != graph.m_Vertices[front_visit.m_vertex_id].m_Neighbors.end(); ++neighbor)
	    {
		sInt_32 neighbor_id = (*neighbor)->m_target->m_id;

		if (visited_Vertices.size() <= front_visit.m_level + 1)
		{
		    sASSERT(visited_Vertices.size() == front_visit.m_level + 1);
		    visited_Vertices.push_back(Visits_umap());
		}
		Visits_umap &next_visited_Vertices = visited_Vertices[front_visit.m_level + 1];

		if (next_visited_Vertices.find(neighbor_id) == next_visited_Vertices.end())
		{
		    if (Conflicts.size() <= front_visit.m_level + 1 || Conflicts[front_visit.m_level + 1].find(neighbor_id) == Conflicts[front_visit.m_level + 1].end())
		    {
			Visit neighbor_visit(front_visit.m_level + 1, neighbor_id, front_visit.m_vertex_id);
			next_visited_Vertices[neighbor_id] = neighbor_visit;

			if (neighbor_id == sink_id)
			{
			    bool non_conflicting_sink = true;

			    for (sInt_32 level = front_visit.m_level + 1; level < Conflicts.size(); ++level)
			    {
				if (Conflicts[level].find(sink_id) != Conflicts[level].end())
				{
				    non_conflicting_sink = false;
				    break;
				}
			    }

			    if (non_conflicting_sink)
			    {
				sInt_32 prev_vertex_id = sink_id;
				Path.resize(front_visit.m_level + 2, -1);
/*
				printf("Path:\n");
				for (sInt_32 i = 0; i <= front_visit.m_level + 1; ++i)
				{
				    for (Visits_umap::const_iterator visit_vertex = visited_Vertices[i].begin(); visit_vertex != visited_Vertices[i].end(); ++visit_vertex)
				    {
					printf("(%d,%d,%d) ", visit_vertex->second.m_level, visit_vertex->second.m_vertex_id, visit_vertex->second.m_previous_id);
				    }
				    printf("\n");
				}
*/			    
				for (sInt_32 i = front_visit.m_level + 1; i >= 0; --i)
				{
				    Path[i] = prev_vertex_id;
				    Visits_umap::const_iterator prev_visit = visited_Vertices[i].find(prev_vertex_id);
				    
				    sASSERT(prev_visit != visited_Vertices[i].end());
				    
				    prev_vertex_id = prev_visit->second.m_previous_id;
				}
				return front_visit.m_level + 2;
			    }
			}
			visit_Queue.push_back(neighbor_visit);
		    }
		}
	    }
	    { /* wait action */
		
		if (visited_Vertices.size() <= front_visit.m_level + 1)
		{
		    sASSERT(visited_Vertices.size() == front_visit.m_level + 1);
		    visited_Vertices.push_back(Visits_umap());
		}		
		Visits_umap &next_visited_Vertices = visited_Vertices[front_visit.m_level + 1];

		if (next_visited_Vertices.find(front_visit.m_vertex_id) == next_visited_Vertices.end())
		{
		    if (Conflicts.size() <= front_visit.m_level + 1 || Conflicts[front_visit.m_level + 1].find(front_visit.m_vertex_id) == Conflicts[front_visit.m_level + 1].end())
		    {
			if (front_visit.m_vertex_id == sink_id)
			{
			    bool non_conflicting_sink = true;

			    for (sInt_32 level = front_visit.m_level + 1; level < Conflicts.size(); ++level)
			    {
				if (Conflicts[level].find(sink_id) != Conflicts[level].end())
				{
				    non_conflicting_sink = false;
				    break;
				}
			    }

			    if (non_conflicting_sink)
			    {
				sInt_32 prev_vertex_id = sink_id;
				Path.resize(front_visit.m_level + 1, -1);
/*
				printf("Path:\n");
				for (sInt_32 i = 0; i <= front_visit.m_level; ++i)
				{
				    for (Visits_umap::const_iterator visit_vertex = visited_Vertices[i].begin(); visit_vertex != visited_Vertices[i].end(); ++visit_vertex)
				    {
					printf("(%d,%d,%d) ", visit_vertex->second.m_level, visit_vertex->second.m_vertex_id, visit_vertex->second.m_previous_id);
				    }
				    printf("\n");
				}
*/
			    
				for (sInt_32 i = front_visit.m_level; i >= 0; --i)
				{
				    Path[i] = prev_vertex_id;
				    Visits_umap::const_iterator prev_visit = visited_Vertices[i].find(prev_vertex_id);
				    
				    sASSERT(prev_visit != visited_Vertices[i].end());
				    
				    prev_vertex_id = prev_visit->second.m_previous_id;
				}
				return front_visit.m_level + 1;
			    }
			}
			
			Visit identity_visit(front_visit.m_level + 1, front_visit.m_vertex_id, front_visit.m_vertex_id);
			next_visited_Vertices[front_visit.m_vertex_id] = identity_visit;
			visit_Queue.push_back(identity_visit);
		    }
		}
	    }
	    visit_Queue.pop_front();
	}
	return -1;
    }


    void sCBS::cast_Occupations(sInt_32 agent_id, sInt_32 prefix_length, const VertexIDs_vector &path, Occupations_vector &space_Occupations) const
    {
	for (sInt_32 i = 0; i < prefix_length; ++i)
	{
	    space_Occupations[i][path[i]] = agent_id;
	}
    }

    
    void sCBS::uncast_Occupations(sInt_32 prefix_length, const VertexIDs_vector &path, Occupations_vector &space_Occupations) const
    {
	for (sInt_32 i = 0; i < prefix_length; ++i)
	{
	    space_Occupations[i].erase(path[i]);
	}		
    }

    
/*----------------------------------------------------------------------------*/

} // namespace boOX
